-module(database).

-behaviour(gen_server).

-export([start_link/1, init/1, handle_cast/2, handle_call/3]).
-export([get_tick/0]).

-record(state, {database_file = none, database = none, tick_start = none}).

start_link(DatabaseFile) ->
  gen_server:start_link({local, ?MODULE}, ?MODULE, [DatabaseFile], []).

open_db(DatabaseFile) ->
  {ok, DB} = sqlite3:open(event_db, [{file, DatabaseFile}]),
  ok = sqlite3:sql_exec(DB, "PRAGMA foreign_keys = ON;"),
  HT = case sqlite3:sql_exec(DB, "SELECT tick FROM match_states;") of
         [{columns, C1}, {rows, R1}] ->
           if length(R1) == 0 -> 0;
              true ->
                T1 = get_column("tick", C1, R1),
                lists:max(T1)
           end;
         _ -> 0
       end,
  HFT = case sqlite3:sql_exec(DB, "SELECT tick FROM finals_states;") of
          [{columns, C2}, {rows, R2}] ->
            if length(R2) == 0 -> 0;
               true ->
                 T2 = get_column("tick", C2, R2),
                 lists:max(T2)
            end;
          _ -> 0
        end,
  HighestTick = lists:max([HT, HFT]),
  TickStart = erlang:monotonic_time(millisecond) - HighestTick,
  {TickStart, DB}.

init([DatabaseFile]) ->
  Exists = filelib:is_regular(DatabaseFile),
  if Exists =:= true ->
       {TickStart, DB} = open_db(DatabaseFile),
       {ok, #state{database_file = DatabaseFile, database = DB, tick_start = TickStart}};
     true ->
       io:format("Failed to load database ~s~n", [DatabaseFile]),
       {ok, #state{database_file = none, database = none}}
  end.

init_db(Database) ->
  ok = sqlite3:create_table(Database, divisions, [{division, integer, [not_null]}],
                            [{primary_key, [division]}]),

  ok = sqlite3:create_table(Database, teams, [{team, text, [not_null]},
                                              {division, integer},
                                              {inspection, blob}],
                            [{primary_key, [team]},
                             {foreign_key, {[division], divisions, [division], ""}}]),

  ok = sqlite3:create_table(Database, matches, [{division, integer},
                                                {type, text, [not_null]},
                                                {number, integer, [not_null]},
                                                {instance, integer, [not_null]}],
                            [{primary_key, [division, type, number, instance]},
                             {check, "type IN ('practice', 'qualification', 'elimination', 'final')"},
                             {foreign_key, {[division], divisions, [division], "ON DELETE CASCADE"}}]),

  ok = sqlite3:create_table(Database, match_teams, [{division, integer, [not_null]},
                                                    {type, text, [not_null]},
                                                    {number, integer, [not_null]},
                                                    {instance, integer, [not_null]},
                                                    {side, text, [not_null]}, 
                                                    {anum, integer, [not_null]},
                                                    {team, text, [not_null]}],
                            [{primary_key, [division, type, number, instance, side, anum]},
                             {check, "side IN ('red', 'blue')"},
                             {check, "anum >= 1 AND anum <= 2"},
                             {foreign_key, {[division, type, number, instance], matches, [division, type, number, instance], "ON DELETE CASCADE"}},
                             {foreign_key, {[team], teams, [team], ""}}]),

  ok = sqlite3:create_table(Database, match_scores, [{division, integer},
                                                     {type, text, [not_null]},
                                                     {number, integer, [not_null]},
                                                     {instance, integer, [not_null]},
                                                     {history, integer, [not_null]},
                                                     {autonomous, integer, [not_null]},
                                                     {red_elevation_1, integer, [not_null]},
                                                     {red_elevation_2, integer, [not_null]},
                                                     {red_goal, integer, [not_null]},
                                                     {red_zone, integer, [not_null]},
                                                     {red_agoal, integer, [not_null]},
                                                     {red_azone, integer, [not_null]},
                                                     {red_1, text},
                                                     {red_2, text},
                                                     {blue_elevation_1, integer, [not_null]},
                                                     {blue_elevation_2, integer, [not_null]},
                                                     {blue_goal, integer, [not_null]},
                                                     {blue_zone, integer, [not_null]},
                                                     {blue_agoal, integer, [not_null]},
                                                     {blue_azone, integer, [not_null]},
                                                     {blue_1, text},
                                                     {blue_2, text}],
                            [{primary_key, [division, type, number, instance, history]},
                             {check, "red_1 IS NULL OR red_1 IN ('dq', 'no show')"},
                             {check, "red_2 IS NULL OR red_2 IN ('dq', 'no show')"},
                             {check, "blue_1 IS NULL OR blue_1 IN ('dq', 'no show')"},
                             {check, "blue_2 IS NULL OR blue_2 IN ('dq', 'no show')"},
                             {foreign_key, {[division, type, number, instance], matches, [division, type, number, instance], "ON DELETE CASCADE"}}]),

  ok = sqlite3:create_table(Database, match_states, [{division, integer},
                                                     {type, text, [not_null]},
                                                     {number, integer, [not_null]},
                                                     {instance, integer, [not_null]},
                                                     {tick, integer, [not_null]},
                                                     {state, text, [not_null]}],
                            [{primary_key, [division, type, number, instance, tick]},
                             {foreign_key, {[division, type, number, instance], matches, [division, type, number, instance], "ON DELETE CASCADE"}}]),

  ok = sqlite3:create_table(Database, alliances, [{division, integer, [not_null]},
                                                  {number, integer, [not_null]},
                                                  {team1, text, [not_null]},
                                                  {team2, text, [not_null]}],
                            [{primary_key, [division, number]},
                             {foreign_key, {[team1], teams, [team], "ON DELETE CASCADE"}},
                             {foreign_key, {[team2], teams, [team], "ON DELETE CASCADE"}}]),

  ok = sqlite3:create_table(Database, rankings, [{division, integer, [not_null]},
                                                 {rank, integer, [not_null]},
                                                 {team, text, [not_null]}],
                            [{primary_key, [division, rank]},
                             {foreign_key, {[division], divisions, [division], "ON DELETE CASCADE"}},
                             {foreign_key, {[team], teams, [team], "ON DELETE CASCADE"}}]),

  ok = sqlite3:create_table(Database, skills_scores, [{team, text, [not_null]},
                                                      {type, text, [not_null]},
                                                      {attempt, integer, [not_null]},
                                                      {elevation, integer, [not_null]},
                                                      {goal, integer, [not_null]},
                                                      {zone, integer, [not_null]},
                                                      {agoal, integer, [not_null]},
                                                      {azone, integer, [not_null]}],
                            [{primary_key, [team, type, attempt]},
                             {check, "type IN ('driver', 'autonomous')"},
                             {foreign_key, {[team], teams, [team], ""}}]),
  ok.

first_error([]) -> ok;
first_error([ok | Rest]) -> first_error(Rest);
first_error([{rowid, _} | Rest]) -> first_error(Rest);
first_error([Error | _]) -> Error.

delete_teams(Database, Teams) ->
  SQL = lists:append(["BEGIN; ",
                      lists:append([io_lib:format("DELETE FROM teams WHERE team = ~p; ",
                                                  [Team]) || Team <- Teams]),
                      "COMMIT;"]),
  first_error(sqlite3:sql_exec_script(Database, SQL)).

add_teams(Database, Teams) ->
  SQL = lists:append(["BEGIN; ",
                      lists:append([io_lib:format("INSERT INTO teams(team) VALUES('~s') ON CONFLICT(team) DO NOTHING; ",
                                                  [Team]) || Team <- Teams]),
                      "COMMIT;"]),
  first_error(sqlite3:sql_exec_script(Database, SQL)).

assign_divisions(Database, Teams) ->
  SQL = lists:append(["BEGIN; ",
                      lists:append([io_lib:format("UPDATE teams SET division = ~p WHERE team = '~s'; ", [Division, Team]) || {Team, Division} <- Teams]),
                      "COMMIT;"]),
  first_error(sqlite3:sql_exec_script(Database, SQL)).

write_division_matches(Database, Division, Type, Matches) ->
  Round = atom_to_list(Type),
  SQL = lists:append(["BEGIN; ",
                      io_lib:format("DELETE FROM matches WHERE division = ~p AND type = '~s';",
                                    [Division, Round]),
                      lists:append([io_lib:format(
                                      "INSERT INTO matches(division, type, number, instance) VALUES (~p, '~s', ~p, 1);
                                       INSERT INTO match_teams(division, type, number, instance, side, anum, team) VALUES (~p, '~s', ~p, 1, 'red',  1, '~s');
                                       INSERT INTO match_teams(division, type, number, instance, side, anum, team) VALUES (~p, '~s', ~p, 1, 'red',  2, '~s');
                                       INSERT INTO match_teams(division, type, number, instance, side, anum, team) VALUES (~p, '~s', ~p, 1, 'blue', 1, '~s');
                                       INSERT INTO match_teams(division, type, number, instance, side, anum, team) VALUES (~p, '~s', ~p, 1, 'blue', 2, '~s');",
                                      [Division, Round, N, Division, Round, N, B1, Division, Round, N, B2, Division, Round, N, R1, Division, Round, N, R2])
                                    || {N, [B1, B2, R1, R2]} <- lists:enumerate(Matches)]),
                      "COMMIT;"]),
  first_error(sqlite3:sql_exec_script(Database, SQL)).

index_of(Item, List) -> index_of(Item, List, 1).
index_of(Item, [Element | _], N) when Item == Element -> N;
index_of(Item, [_ | List], N) -> index_of(Item, List, N+1).

first_empty([Num1, Num2 | Nums]) when Num2 == (Num1 + 1) -> first_empty([Num2 | Nums]);
first_empty([Num | _]) -> Num + 1.

get_column(_, [], _) -> [];
get_column(_, _, []) -> [];
get_column(Name, Columns, Rows) ->
  Index = index_of(Name, Columns),
  [element(Index, Row) || Row <- Rows].

handle_call(get_tick_start, _, State) ->
  {reply, State#state.tick_start, State};
handle_call({new_db, DatabaseFile}, _, State) ->
  ok = if State#state.database =:= none -> ok;
          true -> sqlite3:close(State#state.database)
       end,
  Exists = filelib:is_regular(DatabaseFile),
  ok = if Exists -> file:delete(DatabaseFile);
          true -> ok
       end,
  {TickStart, Database} = open_db(DatabaseFile),
  {reply, init_db(Database), State#state{database_file = DatabaseFile, database = Database, tick_start = TickStart}};
handle_call({load_db, DatabaseFile}, _, State) ->
  true = filelib:is_regular(DatabaseFile),
  {TickStart, Database} = if DatabaseFile == State#state.database_file ->
                               {State#state.tick_start, State#state.database};
                             true ->
                               ok = if State#state.database =:= none -> ok;
                                       true -> sqlite3:close(State#state.database)
                                    end,
                               open_db(DatabaseFile)
                          end,
  {reply, ok, State#state{database_file = DatabaseFile, database = Database, tick_start = TickStart}};
handle_call(close_db, _, State) ->
  if State#state.database =:= none -> {reply, ok, State};
     true ->
       ok = sqlite3:close(State#state.database),
       {reply, ok, State#state{database = none, database_file = none}}
  end;

handle_call({add_teams, Teams}, _, State) ->
  ok = add_teams(State#state.database, Teams),
  {reply, ok, State};
handle_call({assign_divisions, Teams}, _, State) ->
  ok = assign_divisions(State#state.database, Teams),
  {reply, ok, State};
handle_call(split_teams, _, State) ->
  Teams = [X||{_,X} <- lists:sort([{rand:uniform(), N} || N <- get_teams(State#state.database)])],
  [{columns, _}, {rows, Divisions}] = sqlite3:sql_exec(State#state.database, "SELECT division FROM divisions;"),
  Assignments = [{Team, (I rem length(Divisions)) + 1} || {I, Team} <- lists:enumerate(Teams)],
  ok = assign_divisions(State#state.database, Assignments),
  {reply, ok, State};
handle_call({delete_teams, Removed}, _, State) ->
  ok = delete_teams(State#state.database, Removed),
  {reply, ok, State};

handle_call({match_score, Division, Round, Number, Instance, Red, Blue, Score}, _, State) ->
  [{columns, Columns}, {rows, Rows}] = sqlite3:sql_exec(State#state.database,
                                                        io_lib:format(
                                                          "SELECT history FROM match_scores WHERE Division = ~p AND type = '~s' AND number = ~p AND instance = ~p;",
                                                          [Division, atom_to_list(Round), Number, Instance])),
  Instances = get_column("history", Columns, Rows),
  NextInstance = if length(Instances) == 0 -> 1;
                    true -> lists:max(Instances) + 1
                 end,
  ok = first_error(sqlite3:sql_exec_script(State#state.database, lists:append(
                                                                   [io_lib:format("INSERT INTO match_scores(division, type, number, instance, history, red, blue, score) VALUES(~p, '~s', ~p, ~p, ~p, ~p, ~p, '~s');", [Division, atom_to_list(Round), Number, Instance, NextInstance, Red, Blue, Score]),
                                                                    io_lib:format("INSERT INTO match_states(division, type, number, instance, tick, state) VALUES(~p, '~s', ~p, ~p, ~p, '~s');", [Division, atom_to_list(Round), Number, Instance, get_tick(State), "scored"])]))),
  {reply, ok, State};
handle_call({match_state, Division, Round, Number, Instance, MatchState}, _, State) ->
  {rowid, _} = sqlite3:sql_exec(State#state.database,
                                io_lib:format("INSERT INTO match_states(division, type, number, instance, tick, state) VALUES(~p, '~s', ~p, ~p, ~p, '~s');",
                                              [Division, atom_to_list(Round), Number, Instance, get_tick(State), MatchState])),
  {reply, ok, State};

handle_call(add_division, _, State) ->
  [{columns, Columns}, {rows, Rows}] = sqlite3:sql_exec(State#state.database, "SELECT division FROM divisions;"),
  Divisions = lists:sort(get_column("division", Columns, Rows)),
  NextDivision = first_empty([0 | Divisions]),
  {rowid, _} = sqlite3:sql_exec(State#state.database,
                                io_lib:format("INSERT INTO divisions(division) VALUES(~p);",
                                              [NextDivision])),
  {reply, ok, State};
handle_call({delete_division, Division}, _, State) ->
  ok = first_error(sqlite3:sql_exec_script(State#state.database, io_lib:format("UPDATE teams SET division = NULL WHERE division = ~p; DELETE FROM divisions WHERE division = ~p;", [Division, Division]))),
  {reply, ok, State};
handle_call({generate_division, Division, Round, Size, Seed}, _, State) ->
  Teams = get_div_teams(State#state.database, Division),
  io:fwrite("Generating division ~p/~s with teams ~p, size ~p, and seed ~p~n", [Division, Round, Teams, Size, Seed]),
  Matches = schedule:create(Seed, Size, Teams),
  ok = write_division_matches(State#state.database, Division, Round, Matches),
  {reply, ok, State}.

get_div_teams(Database, Division) ->
  [{columns, Columns}, {rows, Rows}] = sqlite3:sql_exec(Database, io_lib:format("SELECT team FROM teams WHERE division = ~p;", [Division])),
  get_column("team", Columns, Rows).

get_teams(Database) ->
  [{columns, Columns}, {rows, Rows}] = sqlite3:sql_exec(Database, "SELECT team FROM teams;"),
  get_column("team", Columns, Rows).

handle_cast(_, State) ->
  {noreply, State}.

get_tick(State) -> erlang:monotonic_time(millisecond) - State#state.tick_start.
get_tick() ->
  erlang:monotonic_time(millisecond) - gen_server:call(?MODULE, get_tick_start).
